//////////////////////////////////////////////////////////////////////////
//                                                                      //
// fmodf.cpp -                                                          //
//                                                                      //
//                              //
//        .                          //
//      (     Y)
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:07p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

#ifdef __cplusplus
extern "C" {
#endif

// fmodf --
//
//         .  
//
float fmodf(float X, float Y)
{
    /*
        fmod, fmodf, fmodl
        double fmod(double x, double y);
        float fmod(float x, float y); [C++ only]
        long double fmod(long double x, long double y); [C++ only]
        float fmodf(float x, float y); [optional]
        long double fmodl(long double x, long double y); [optional] 

        The function returns the remainder of x/y, which is defined as follows:

        If y is zero, the function either reports a domain error or simply returns zero. 
        Otherwise, if 0 <= x, the value is x - i*y for some integer i such that:
        0 <= i*|y| <= x < (i + 1)*|y| 
        Otherwise, x < 0 and the value is x - i*y for some integer i such that:
        i*|y| <= x < (i + 1)*|y| <= 0 
    */
	int Xm, Ym;                     //   
    int tmp;                        //     Xm
	int Xe, Xs;                     //      
	int Ye, Ys;                     //      
    int DeltaE;                     //   
 	int sc;

    //    
    DestructFloat(X, Xm, Xe, Xs, sc);

    //    
	DestructFloat(Y, Ym, Ye, Ys, sc);

    if ( Xe == 0x80 ){ // X = inf, NAN
        return ConstructFloat(1, 0x80, 0, 0);
    }

    if ( Ye == 0x80 && Ym ){    // NAN
        return Ym;
    }

    if ( Ye == -0x7f && (!Ym) ){  //  x/0
    	return ConstructFloat(1, 0x80, 0, 0);   // NAN
	}

    if (Xe > Ye+24 || Xe < -103 ){  //  103 = 0x7f - 24 
    	return ConstructFloat(0, -0x7f, Xs, 0);   // 0
	}

    if ( Ye > Xe ){     //       -  
        return X;
    }

    DeltaE = Xe - Ye;

    //      
    Xm <<= 7;
    Ym <<= 7;
    tmp = Xm;
    do{
        if ( (Xm-=Ym)>0 ){
            Xm <<=1;
            tmp = Xm;
            Xe--;
        }
        else{
            Xm = tmp;
            Ym >>=1;
        }
    }while (DeltaE--);

    while (!(Xm & 0x40000000)){
        Xm <<=1;
        Xe--;
    }
    Xm >>= 7;

    //   float
	return ConstructFloat(Xm, Xe, Xs, 0);
}

#ifdef __cplusplus
}
#endif
