#ifndef	_MALLOC_INTERNAL
#define _MALLOC_INTERNAL
#include <malloc.h>
#endif
#include <assert.h>

/* Debugging hook for free.  */
void (*__free_hook) __P ((__ptr_t __ptr));

/* List of blocks allocated by memalign.  */
/*struct alignlist *_aligned_blocks = NULL;*/

/* Return memory to the heap.
   Like `free' but don't call a __free_hook if there is one.  */
void __free_internal (__ptr_t ptr,_heap_control_block *cb)
{
  int type;
  __malloc_size_t block, blocks;
  register __malloc_size_t i;
  struct list *prev, *next;

//  assert(ptr!=NULL);

  block = BLOCK (ptr,cb);

  type = cb->_heapinfo[block].busy.type;
  switch (type)
    {
    case 0:
      /* Get as many statistics as early as we can.  */
      --cb->_chunks_used;
      cb->_bytes_used -= cb->_heapinfo[block].busy.info.size * BLOCKSIZE;
      cb->_bytes_free += cb->_heapinfo[block].busy.info.size * BLOCKSIZE;

      /* Find the free cluster previous to this one in the free list.
	 Start searching at the last block referenced; this may benefit
	 programs with locality of allocation.  */
      i = cb->_heapindex;
      if (i > block)
	while (i > block)
	  i = cb->_heapinfo[i].free.prev;
      else
	{
	  do
	    i = cb->_heapinfo[i].free.next;
	  while (i > 0 && i < block);
	  i = cb->_heapinfo[i].free.prev;
	}

      /* Determine how to link this block into the free list.  */
      if (block == i + cb->_heapinfo[i].free.size)
	{
	  /* Coalesce this block with its predecessor.  */
	  cb->_heapinfo[i].free.size += cb->_heapinfo[block].busy.info.size;
	  block = i;
	}
      else
	{
	  /* Really link this block back into the free list.  */
	  cb->_heapinfo[block].free.size = cb->_heapinfo[block].busy.info.size;
	  cb->_heapinfo[block].free.next = cb->_heapinfo[i].free.next;
	  cb->_heapinfo[block].free.prev = i;
	  cb->_heapinfo[i].free.next = block;
	  cb->_heapinfo[cb->_heapinfo[block].free.next].free.prev = block;
	  ++cb->_chunks_free;
	}

      /* Now that the block is linked in, see if we can coalesce it
	 with its successor (by deleting its successor from the list
	 and adding in its size).  */
      if (block + cb->_heapinfo[block].free.size == cb->_heapinfo[block].free.next)
	{
	  cb->_heapinfo[block].free.size
	    += cb->_heapinfo[cb->_heapinfo[block].free.next].free.size;
	  cb->_heapinfo[block].free.next
	    = cb->_heapinfo[cb->_heapinfo[block].free.next].free.next;
	  cb->_heapinfo[cb->_heapinfo[block].free.next].free.prev = block;
	  --cb->_chunks_free;
	}

      /* Now see if we can return stuff to the system.  */
      blocks = cb->_heapinfo[block].free.size;
      if (blocks >= FINAL_FREE_BLOCKS && block + blocks == cb->_heaplimit
	  && cb->__morecore(0) == ADDRESS (block + blocks,cb))
	{
	  register __malloc_size_t bytes = blocks * BLOCKSIZE;
	  cb->_heaplimit -= blocks;
	  cb->__morecore(-bytes);
	  cb->_heapinfo[cb->_heapinfo[block].free.prev].free.next
	    = cb->_heapinfo[block].free.next;
	  cb->_heapinfo[cb->_heapinfo[block].free.next].free.prev
	    = cb->_heapinfo[block].free.prev;
	  block = cb->_heapinfo[block].free.prev;
	  --cb->_chunks_free;
	  cb->_bytes_free -= bytes;
	}

      /* Set the next search to begin at this block.  */
      cb->_heapindex = block;
      break;

    default:
      /* Do some of the statistics.  */
		int lesOptShift1 = 1 << type;
		int lesOptShift2 = BLOCKSIZE >> type;

      --cb->_chunks_used;
      cb->_bytes_used -= lesOptShift1;
      ++cb->_chunks_free;
      cb->_bytes_free += lesOptShift1;

      /* Get the address of the first free fragment in this block.  */
      prev = (struct list *) ((char *) ADDRESS (block,cb) +
			   (cb->_heapinfo[block].busy.info.frag.first << type));

      if (cb->_heapinfo[block].busy.info.frag.nfree == (lesOptShift2) - 1)
	{
	  /* If all fragments of this block are free, remove them
	     from the fragment list and free the whole block.  */
	  next = prev;
	  for (i = 1; i < (__malloc_size_t) (lesOptShift2); ++i)
	    next = next->next;
	  prev->prev->next = next;
	  if (next != NULL)
	    next->prev = prev->prev;
	  cb->_heapinfo[block].busy.type = 0;
	  cb->_heapinfo[block].busy.info.size = 1;

	  /* Keep the statistics accurate.  */
	  ++cb->_chunks_used;
	  cb->_bytes_used += BLOCKSIZE;
	  cb->_chunks_free -= lesOptShift2;
	  cb->_bytes_free -= BLOCKSIZE;

	  free (ADDRESS (block,cb));
	}
      else if (cb->_heapinfo[block].busy.info.frag.nfree != 0)
	{
	  /* If some fragments of this block are free, link this
	     fragment into the fragment list after the first free
	     fragment of this block. */
	  next = (struct list *) ptr;
	  next->next = prev->next;
	  next->prev = prev;
	  prev->next = next;
	  if (next->next != NULL)
	    next->next->prev = next;
	  ++cb->_heapinfo[block].busy.info.frag.nfree;
	}
      else
	{
	  /* No fragments of this block are free, so link this
	     fragment into the fragment list and announce that
	     it is the first free fragment of this block. */
	  prev = (struct list *) ptr;
	  cb->_heapinfo[block].busy.info.frag.nfree = 1;
	  cb->_heapinfo[block].busy.info.frag.first = (unsigned int)
	    ((unsigned int) ((char *) ptr - (char *) NULL)
	     % lesOptShift2);
	  prev->next = cb->_fraghead[type].next;
	  prev->prev = &(cb->_fraghead[type]);
	  prev->prev->next = prev;
	  if (prev->next != NULL)
	    prev->next->prev = prev;
	}
      break;
    }
}

/* Return memory to the heap.  */
void free0 (__ptr_t ptr)
{
  if (ptr == NULL) return;
#ifdef _HOME_DEBUG
  printf("<f %p> ",ptr);
#endif
  assert(__IN_HEAP0(ptr));
  __free_internal (ptr,&_hcb0);
}

void free1 (__ptr_t ptr)
{
 if (ptr == NULL)
    return;
#ifdef _HOME_DEBUG
 printf("<f %p> ",ptr);
#endif
 assert(__IN_HEAP1(ptr));
 __free_internal (ptr,&_hcb1);
}

void free2 (__ptr_t ptr)
{
  if (ptr == NULL) return;
#ifdef _HOME_DEBUG
  printf("<f %p> ",ptr);
#endif
  assert(__IN_HEAP2(ptr));
  __free_internal (ptr,&_hcb2);
}

void free3 (__ptr_t ptr)
{
  if (ptr == NULL) return;
#ifdef _HOME_DEBUG
  printf("<f %p> ",ptr);
#endif
  assert(__IN_HEAP3(ptr));
  __free_internal (ptr,&_hcb3);
}

void free (__ptr_t ptr)
{
  if (ptr == NULL) return;
#ifdef _HOME_DEBUG
  printf("<f %p> ",ptr);
#endif
  if (__IN_HEAP0(ptr))
  {
     __free_internal (ptr,&_hcb0);
     return;
  }
  else if (__IN_HEAP1(ptr))
  {
     __free_internal (ptr,&_hcb1);
     return;
  }
  else if (__IN_HEAP2(ptr))
  {
     __free_internal (ptr,&_hcb2);
     return;
  }
  assert(__IN_HEAP3(ptr));
  __free_internal (ptr,&_hcb3);
}
