//////////////////////////////////////////////////////////////////////////
//                                                                      //
//               NeuroMatrix(r) C runtime library                       //
//                                                                      //
//  printf.c                                                            //
//  ,  .                                      //
//                                                                      //
// Copyright (c) 2004-2008 RC Module                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
// $Revision:: 2    $      $Date:: 3.10.11 15:30     $                  //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include"host_duties.h"
#include <stddef.h>
#include <stdio.h>
#include <stdarg.h>

extern "C"
{
	// Barrier sync with host PC with array exchange
	int ncl_hostSyncArray(
		int value,        // Sync value
		void *outAddress, // Sended array address (can be NULL)
		size_t outLen,    // Sended array length (can be 0)
		void **inAddress, // Received array address pointer (can be NULL)
		size_t *inLen);   // Received array size pointer (can be NULL)

    int vsprintf( char *dst, const char* format, va_list args );
};

const int printfBufSize = 0x4000;  //     0x40000 max
int printfBufferLimit = 0x2000;    //      

#pragma data_section ".printf_section"
int printfBuffer1[ printfBufSize ];
int printfBuffer2[ printfBufSize ];
#pragma default_data_section

int* currBuffer= printfBuffer1;
int inBuffer= 0;

void swapBuffers()
{ 
    currBuffer = (printfBuffer1-currBuffer)+printfBuffer2;
}

extern "C" int printf( const char* format,...)
{
    va_list argptr;
    va_start(argptr, format);
    int r= vsprintf( (char*)(currBuffer)+inBuffer, format, argptr);
    inBuffer+= r;
    if ( inBuffer>= printfBufferLimit ){
        ncl_hostSyncArray( NMSYNC_SERVICE_PRINTF, currBuffer, 
                                            inBuffer, 0, 0 );
        inBuffer= 0;
        swapBuffers();
    }
    va_end( argptr );
    return r;
}

class __FlushPrintf_dummy_type
{
public:
    ~__FlushPrintf_dummy_type()
    {
        if ( inBuffer ){
            ncl_hostSyncArray( NMSYNC_SERVICE_PRINTF, currBuffer, 
                                                inBuffer, 0, 0 );
        }
    }
};

//         
__FlushPrintf_dummy_type __flushPrintf_dummy_object;